/* Controlling SW for Arduino Nano, Semi-automatic antenna tuner:
 *  https://www.youtube.com/watch?v=gbERgJ4-GSk
 *  by Petr, OK1FIG, ok1fig@seznam.cz
 *  Last update: 1.3.2017
 */

/*
 * Bands: 
 * 1.8
 * 3.5
 * 3.7
 * 7
 * 10
 * 14
 * 18
 * 21
 * 24
 * 28
 * 50
 * 
 */


#include <EEPROM.h>

// if DEFINE is on, the button doesn't work (as it is used by Serial for communication)
#define _DEBUG

const int ENC_PIN_A = 3;
const int ENC_PIN_B = 4;
const int MIN_ENC_VAL = 0;
const int RELAY_DELAY = 5; // milliseconds

const int CHANGE_BAND_DELAY = 500; // mS

const int MAX_ENC_VAL = 255; // Valid for all relays resolution
//const int MAX_ENC_VAL = 248; // Valid for 5 relays resolution
//const int MAX_ENC_VAL = 240; // Valid for 4 relays resolution
//const int MAX_ENC_VAL = 224; // Valid for 3 relays resolution

const byte BANDS = 11; // Number of supported bands

// measured value for each position of the band switch, sorted ascending:
int BAND_LIMITS[1 + BANDS] = {0, 156, 194, 280, 361, 443, 516, 624, 686, 720, 751, 1023};



#ifdef DEBUG
String debug = "";
String prev_debug = "*";
#endif

// 
int prev_0;
int prev_1; 

byte CurrBand; 
byte CurrBandPrev = 255;
 
long btn_toggle_hi_lo_imp_down = 0; // Debounce toggling button
long btn_toggle_hi_lo_imp_down2 = 0; // Debounce toggling button
long band_changed = 0; // Delay band switching


int EncoderPos[BANDS][2]; // 0 - conds, 1 - coils 
byte EncoderIdx = 0; // conds

int EncoderLastPos[BANDS][2] = {LOW, LOW};
byte HiLoConfig[BANDS];

int n = LOW;

long ms;

int i, j, k;

// For the alternative configuration. Keep pressed during startup to switch to the other configuration
int eeprom_offset = 0;

bool virgin = true;


/***************************************************************************************************************************************************/


void setup() 
  {

  #ifdef DEBUG
  Serial.begin(9600); // If this is enabled then setting D1 doesn't work
  #endif
    
  pinMode (ENC_PIN_A, INPUT_PULLUP);
  pinMode (ENC_PIN_B, INPUT_PULLUP);

  // Prepare pins that will control the relays for condensers:
  for (i = 0; i < 8; i++)
    pinMode (5 + i, OUTPUT);

  // Prepare pins that will control the relays for coils:
  pinMode (A0, OUTPUT);
  pinMode (A1, OUTPUT);
  pinMode (A2, OUTPUT);
  pinMode (A3, OUTPUT);
  pinMode (A4, OUTPUT);
  pinMode (A5, OUTPUT);
  pinMode(2, OUTPUT); 
  pinMode (13, OUTPUT);


  // Analog A6 is used to detect position of the C/L switching switch:
  pinMode (A6, INPUT);
  
  pinMode(0, INPUT_PULLUP); // Toggle Hi/Lo Imp, a button
  pinMode(1, OUTPUT); // Used to switch Hi/Lo relays
  
  pinMode(A7, INPUT); // Sensing band-switching voltage from Icom

  // Holding the button down during startup will switch to the alternative configuratio:
  if (digitalRead(0) == LOW) // Button toggling Hi/Lo imp kept down during the startup
    {
    btn_toggle_hi_lo_imp_down = -1; 
    eeprom_offset = BANDS * (sizeof(EncoderPos[0][0]) + sizeof(EncoderPos[0][1]) + sizeof(HiLoConfig[0])); // Switch to the alternative configuration. Keep this until it device is switched off.
    Blink();
    #ifdef DEBUG
    Serial.print("\nAlternative configuration selected.\n");
    #endif
    }

  ReadFromEeprom();
  
  // Clear the virgin nonsense read from EEPROM. Set two middle relays for both conds and coils:
  for (i = 0; i < BANDS; i++)
    {
    for (j = 0; j < 2; j++)
      if (EncoderPos[i][j] < MIN_ENC_VAL || EncoderPos[i][j] > MAX_ENC_VAL) 
        {
        #ifdef DEBUG  
        Serial.print("Fixing C or L value, band: ");
        Serial.println(i);
        Serial.print("Value: ");
        Serial.println(EncoderPos[i][j]);
        #endif
        EncoderPos[i][j] = 24;
        }

    
    if (HiLoConfig[i] < 0 || HiLoConfig[i] > 1) 
      {
      HiLoConfig[i] = 0;
      #ifdef DEBUG
      Serial.print("Fixing Hi/Lo imp. value, band: ");
      Serial.println(i);
      #endif
      }
    }


    
  } // setup()

/***************************************************************************************************************************************************/

 void loop() 
  { 
  // Read band-switching voltage:
  i = analogRead(A7);

  #ifdef DEBUG
  debug = " A7 (band volt.): " + String(i); 
  #endif

  for (j = 1; j <= BANDS; j++)
    {
//    if (i < BAND_LIMITS[j])
    if (i < (BAND_LIMITS[j-1] + BAND_LIMITS[j]) / 2)
      {
      //CurrBand = j;
      if (band_changed == 0 && CurrBand != j)
        band_changed = millis();
      else if (millis() - band_changed > CHANGE_BAND_DELAY)
        {
        CurrBand = j - 1;
        band_changed = 0;   
        }
      break;
      }     
    }

  
  #ifdef DEBUG
  debug += " Band idx: " + String(CurrBand); // + CRLF;
  #endif
    
  // Update value of the encoder:
  if (analogRead(A6) < 512) // Button up, changing condensers
    EncoderIdx = 1;
  else 
    EncoderIdx = 0; 

  #ifdef DEBUG
  debug += " C or L (0=conds): " + String(EncoderIdx); 
  #endif
  
  n = digitalRead(ENC_PIN_A);
  if (((EncoderLastPos[CurrBand][EncoderIdx] == LOW) && (n == HIGH)) || (CurrBand != CurrBandPrev)) 
    {
    if (digitalRead(ENC_PIN_B) == LOW) 
      j = EncoderPos[CurrBand][EncoderIdx] - 1;
    else 
      j = EncoderPos[CurrBand][EncoderIdx];
  
    // Evaluate increment for tuning:
    /*
    k = 0; //
    while (j > 0) 
      {
      j = j >> 1;
        k++;
      };
    
    k = k - 5; // Use max 5 relays at a time
    //k = k - 4; // Use max 4 relays at a time
    // k = k - 3; // Use max 3 relays at a time
    
    if (k < 0)
      k = 0;
    k = 1 << k; 
    */
    k = 1; // Valid for all relays

    // On startup and the first pass of loop() the frequency must not change. Therefore set the tuning step to zero:
    if (CurrBand != CurrBandPrev)
      k = 0;

    // Add a calculated value; this keeeps only 5 most important relays in action:      
    if (digitalRead(ENC_PIN_B) == LOW) 
      EncoderPos[CurrBand][EncoderIdx] = EncoderPos[CurrBand][EncoderIdx] - k;
    else 
      EncoderPos[CurrBand][EncoderIdx] = EncoderPos[CurrBand][EncoderIdx] + k;

    // Keep value in limits, don't overflow:
    if (EncoderPos[CurrBand][EncoderIdx] > MAX_ENC_VAL)
      {
      EncoderPos[CurrBand][EncoderIdx] = MAX_ENC_VAL;
      }
    else if (EncoderPos[CurrBand][EncoderIdx] < MIN_ENC_VAL)
      {
      EncoderPos[CurrBand][EncoderIdx] = MIN_ENC_VAL;
      }
    } 
  EncoderLastPos[CurrBand][EncoderIdx] = n;


  #ifndef DEBUG // The button is working wrong anyway as Serial is used, disable it:
  ms = millis();  
  if (digitalRead(0) == LOW)
    {
    if (btn_toggle_hi_lo_imp_down == 0)
      btn_toggle_hi_lo_imp_down = ms;
    else
      if (ms - btn_toggle_hi_lo_imp_down > 800 && btn_toggle_hi_lo_imp_down > 0) // The button kept down for 1 second, or so
        {
        SaveToEeprom(CurrBand);
        
        #ifdef DEBUG
        Serial.write(13); Serial.write(10);
        Serial.print("Saved: "); Serial.write(13); Serial.write(10);
        #endif
        
        // Shortly blink with LEDs to indicate that saving happened:
        Blink();

        #ifdef DEBUG
        Serial.print("Blinked");Serial.write(13); Serial.write(10);
        prev_debug = "";
        #endif
            
        btn_toggle_hi_lo_imp_down = -1; 
        btn_toggle_hi_lo_imp_down2 = millis();
        }
    }
  else
    {
    if (btn_toggle_hi_lo_imp_down > 0)
      {  
      ms = ms - btn_toggle_hi_lo_imp_down; 
      // Toggle Hi/Lo impedance:
      if (ms > 50 && ms < 750) // The button shortly hit
        {
        HiLoConfig[CurrBand] = ! HiLoConfig[CurrBand]; // Toggle
        }
      }
    btn_toggle_hi_lo_imp_down = 0;  
    btn_toggle_hi_lo_imp_down2 = 0;   
    }
  #endif

  if (millis() - btn_toggle_hi_lo_imp_down2 > 3000 && btn_toggle_hi_lo_imp_down2 != 0) 
    {
    /*  
    for (i = 0; i < 8; i++)
      {
      digitalWrite(5 + i, true);      
      delay(50);
      digitalWrite(5 + i, false);      
      }
    */
    setup();
    Blink();  
    btn_toggle_hi_lo_imp_down = 0;  
    btn_toggle_hi_lo_imp_down2 = 0;   
    }

  #ifdef DEBUG
  debug += " Conds: " + String(EncoderPos[CurrBand][0]);
  debug += " Coils: " + String(EncoderPos[CurrBand][1]);
  debug += " Hi/Lo: " + String(HiLoConfig[CurrBand]); 
  #endif

  //  The following section switches the relays. It is in one large "if - else if" block in order only one relay switches at a time, during one cycle.
  //  This should, hopefully, make the switching smoother and less noisy.
  // Set Hi-Lo relays:
  if (digitalRead(1) != bool(HiLoConfig[CurrBand]))  {digitalWrite(1, bool(HiLoConfig[CurrBand])); }
  // Set condensers:
  else if (digitalRead(5) != bool(bitRead(EncoderPos[CurrBand][0], 0)))  {digitalWrite(5,  bool(bitRead(EncoderPos[CurrBand][0], 0))); }
  else if (digitalRead(6) != bool(bitRead(EncoderPos[CurrBand][0], 1)))  {digitalWrite(6,  bool(bitRead(EncoderPos[CurrBand][0], 1))); }
  else if (digitalRead(7) != bool(bitRead(EncoderPos[CurrBand][0], 2)))  {digitalWrite(7,  bool(bitRead(EncoderPos[CurrBand][0], 2))); }
  else if (digitalRead(8) != bool(bitRead(EncoderPos[CurrBand][0], 3)))  {digitalWrite(8,  bool(bitRead(EncoderPos[CurrBand][0], 3))); }
  else if (digitalRead(9) != bool(bitRead(EncoderPos[CurrBand][0], 4)))  {digitalWrite(9,  bool(bitRead(EncoderPos[CurrBand][0], 4))); }
  else if (digitalRead(10) != bool(bitRead(EncoderPos[CurrBand][0], 5))) {digitalWrite(10,  bool(bitRead(EncoderPos[CurrBand][0], 5)));}
  else if (digitalRead(11) != bool(bitRead(EncoderPos[CurrBand][0], 6))) {digitalWrite(11,  bool(bitRead(EncoderPos[CurrBand][0], 6)));}
  else if (digitalRead(12) != bool(bitRead(EncoderPos[CurrBand][0], 7))) {digitalWrite(12,  bool(bitRead(EncoderPos[CurrBand][0], 7)));}
  // Set coils:
  else if (digitalRead(13) != ! bool(bitRead(EncoderPos[CurrBand][1], 0))) {digitalWrite(13, ! bool(bitRead(EncoderPos[CurrBand][1], 0))); }
  else if (digitalRead(A0) != ! bool(bitRead(EncoderPos[CurrBand][1], 1))) {digitalWrite(A0, ! bool(bitRead(EncoderPos[CurrBand][1], 1))); }
  else if (digitalRead(A1) != ! bool(bitRead(EncoderPos[CurrBand][1], 2))) {digitalWrite(A1, ! bool(bitRead(EncoderPos[CurrBand][1], 2))); }
  else if (digitalRead(A2) != ! bool(bitRead(EncoderPos[CurrBand][1], 3))) {digitalWrite(A2, ! bool(bitRead(EncoderPos[CurrBand][1], 3))); }
  else if (digitalRead(A3) != ! bool(bitRead(EncoderPos[CurrBand][1], 4))) {digitalWrite(A3, ! bool(bitRead(EncoderPos[CurrBand][1], 4))); }
  else if (digitalRead(A4) != ! bool(bitRead(EncoderPos[CurrBand][1], 5))) {digitalWrite(A4, ! bool(bitRead(EncoderPos[CurrBand][1], 5))); }
  else if (digitalRead(A5) != ! bool(bitRead(EncoderPos[CurrBand][1], 6))) {digitalWrite(A5, ! bool(bitRead(EncoderPos[CurrBand][1], 6))); }
  else if (digitalRead(2) != ! bool(bitRead(EncoderPos[CurrBand][1], 7)))  {digitalWrite(2,  ! bool(bitRead(EncoderPos[CurrBand][1], 7))); }

  // Encoder:
  if (EncoderPos[CurrBand][0] != prev_0 || EncoderPos[CurrBand][1] != prev_1)
    {
    prev_0 = EncoderPos[CurrBand][0];
    prev_1 = EncoderPos[CurrBand][1];
    }


//  virgin_start = false;
  CurrBandPrev = CurrBand;

  #ifdef DEBUG  
  if (debug != prev_debug)
    {
    Serial.println(debug);
    prev_debug = debug;
    }
  #endif
  }  // loop()



/***************************************************************************************************************************************************/
void  Blink()
  {
  int i;
  for (i = 0; i < 2; i++)
    {  
    digitalWrite(1, true); delay(30);        
    digitalWrite(1, false); delay(30);        
    } 
  }
/***************************************************************************************************************************************************/


/***************************************************************************************************************************************************/
/* Save data for the current band into EEPROM
 *  
 */
void  SaveToEeprom(int BandIdx)
  {
  int j;
  j = sizeof(EncoderPos[0][0]) + sizeof(EncoderPos[0][1]) + sizeof(HiLoConfig[0]);

  if (BandIdx >= 0 && BandIdx < BANDS)
    {
    EEPROM.put(eeprom_offset + BandIdx * j, EncoderPos[BandIdx][0]);  
    EEPROM.put(eeprom_offset + BandIdx * j + sizeof(EncoderPos[BandIdx][0]), EncoderPos[BandIdx][1]);  
    EEPROM.put(eeprom_offset + BandIdx * j + sizeof(EncoderPos[BandIdx][0]) + sizeof(EncoderPos[BandIdx][1]), HiLoConfig[BandIdx]);  

    Serial.print("Store to EEPROM: "); 
    Serial.print("Band: "); Serial.print(BandIdx);
    Serial.print("/"); 
    Serial.print(EncoderPos[BandIdx][0]);
    Serial.print("/"); 
    Serial.print(EncoderPos[BandIdx][1]);
    Serial.print("/"); 
    Serial.println(HiLoConfig[BandIdx]);
    }
  }
/***************************************************************************************************************************************************/


/***************************************************************************************************************************************************/
/* Read data for all bands from EEPROM.
 *  
 */
void  ReadFromEeprom()
  {
  int i, j, offset;
  
  j = sizeof(EncoderPos[0][0]) + sizeof(EncoderPos[0][1]) + sizeof(HiLoConfig[0]);
  
  for (i = 0; i < BANDS; i++)
    {
    EEPROM.get(eeprom_offset + i * j, EncoderPos[i][0]);
    EEPROM.get(eeprom_offset + i * j + sizeof(EncoderPos[i][0]), EncoderPos[i][1]);
    EEPROM.get(eeprom_offset + i * j + sizeof(EncoderPos[i][0]) + sizeof(EncoderPos[i][0]), HiLoConfig[i]);

    Serial.print("Read from EEPROM, band:"); 
    Serial.print(i);
    Serial.print("/"); 
    Serial.print(EncoderPos[i][0]);
    Serial.print("/"); 
    Serial.print(EncoderPos[i][1]);
    Serial.print("/"); 
    Serial.println(HiLoConfig[i]);
    }
  }
/***************************************************************************************************************************************************/


